from textwrap import indent

from requests.utils import requote_uri
from tabulate import tabulate
from typing import List

from janis_core import Workflow, WorkflowMetadata
from .utils import (
    prepare_byline,
    format_rst_link,
    get_tool_url,
    prepare_quickstart,
    prepare_container_warning_for_workflow,
)

SHOW_WORKFLOW_IMAGE = False


def prepare_workflow_page(workflow: Workflow, versions: List[str]):
    if not workflow:
        return None

    metadata: WorkflowMetadata = workflow.bind_metadata() or workflow.metadata

    if not workflow.friendly_name():
        raise Exception(
            f"Tool '{type(workflow).__name__}' ({workflow.id()}) did not provide the required 'friendly_name' for the docs"
        )

    fn = workflow.friendly_name() if workflow.friendly_name() else workflow.id()
    en = f" ({workflow.id()})" if fn != workflow.id() else ""
    tn = fn + en

    onelinedescription = prepare_byline(
        workflow.id(), metadata.short_documentation, metadata.contributors, versions
    )

    citation = "\n\n".join([el for el in [metadata.citation, metadata.doi] if el])

    formatted_url = (
        format_rst_link(metadata.documentationUrl, metadata.documentationUrl)
        if metadata.documentationUrl
        else "*No URL to the documentation was provided*"
    )

    toolmetadata = [
        ("ID", f"``{workflow.id()}``"),
        ("URL", formatted_url),
        ("Versions", ", ".join(str(s) for s in versions[::-1]) if versions else ""),
        ("Authors", ", ".join(metadata.contributors)),
        ("Citations", citation),
        ("Created", str(metadata.dateCreated)),
        ("Updated", str(metadata.dateUpdated)),
    ]

    embeddedtoolsraw = {
        f"{s.tool.id()}/{s.tool.version()}": s.tool
        for s in workflow.step_nodes.values()
    }
    embeddedtools = tabulate(
        [
            [tool.friendly_name(), f"``{key}``"]
            for key, tool in embeddedtoolsraw.items()
        ],
        tablefmt="rst",
    )

    input_headers = ["name", "type", "documentation"]

    required_input_tuples = [
        [i.id(), i.intype.id(), i.doc.doc if i.doc else ""]
        for i in workflow.tool_inputs()
        if not i.intype.optional
    ]
    optional_input_tuples = [
        [i.id(), i.intype.id(), i.doc.doc if i.doc else ""]
        for i in workflow.tool_inputs()
        if i.intype.optional
    ]

    formatted_inputs = tabulate(
        required_input_tuples + optional_input_tuples, input_headers, tablefmt="rst"
    )

    formatted_toolversions_array = []
    formatted_toolincludes_array = []
    for v in versions:
        link = get_tool_url(workflow.id(), v)
        formatted_toolincludes_array.append(".. include:: " + link)
        if v == workflow.version():
            formatted_toolversions_array.append(
                f"- {v} (current)"
            )  # + format_rst_link(v + " (current)", link))
        else:
            formatted_toolversions_array.append(
                "- " + format_rst_link(v, link + ".html")
            )

    output_headers = ["name", "type", "documentation"]
    output_tuples = [
        [o.id(), o.outtype.id(), o.doc.doc] for o in workflow.tool_outputs()
    ]
    formatted_outputs = tabulate(output_tuples, output_headers, tablefmt="rst")

    cwl = workflow.translate("cwl", to_console=False, allow_empty_container=True)[0]
    wdl = workflow.translate("wdl", to_console=False, allow_empty_container=True)[0]

    tool_prov = ""
    if workflow.tool_provider() is None:
        print("Tool :" + workflow.id() + " has no company")
    else:
        tool_prov = "." + workflow.tool_provider().lower()

    workflow_image = requote_uri(workflow.versioned_id()) + ".dot.png"

    nl = "\n"

    return f"""\
:orphan:

{fn}
{"=" * len(tn)}

{onelinedescription}

{metadata.documentation if metadata.documentation else "No documentation was provided: " + format_rst_link(
    "contribute one", f"https://github.com/PMCC-BioinformaticsCore/janis-{workflow.tool_module()}")}

{prepare_container_warning_for_workflow(workflow)}
{prepare_quickstart(workflow)}

Information
------------

URL: {formatted_url}

{nl.join(f":{key}: {value}" for key, value in toolmetadata)}



Outputs
-----------

{formatted_outputs}


Workflow
--------

.. image:: {workflow_image}

Embedded Tools
***************

{embeddedtools}



Additional configuration (inputs)
---------------------------------

{formatted_inputs}

Workflow Description Language
------------------------------

.. code-block:: text

{indent(wdl, "   ")}

Common Workflow Language
-------------------------

.. code-block:: text

{indent(cwl, "   ")}
"""



import janis_core as j
from janis.tools import Echo

w = j.WorkflowBuilder("workflowId")

w.input("input_id", j.String, default="Hello, World!")
w.step("step_id", Echo(inp=w.input_id))
w.output("output_id", source=w.step_id.out)

# Will print the CWL, input file and relevant tools to the console
if __name__ == "__main__":
    w.translate("cwl", to_disk=False)  # or "wdl"



from textwrap import indent

from janis_core.workflow.workflow import InputNode
from tabulate import tabulate
from typing import List, Dict
from datetime import date
from distutils.version import StrictVersion

from janis_core import (
    Tool,
    Metadata,
    DataType,
    Array,
    File,
    String,
    Int,
    Float,
    Boolean,
    InputQualityType,
    CommandTool,
    Workflow,
    WorkflowBase,
)
from janis_core.translations import CwlTranslator
from requests.utils import requote_uri


class TocObject:
    def __init__(self, title, description, url):
        self.title = title
        self.description = description
        self.url = url


class NestedDictionaryTypeException(Exception):
    def __init__(self, key, error_type, keys=None):
        keychain = (" for keys: " + ".".join(keys)) if keys else ""
        super(NestedDictionaryTypeException, self).__init__(
            f"Incorrect type '{error_type}' for key '{key}'{keychain}"
        )
        self.keys = keys
        self.key = key
        self.error_type = error_type


def format_rst_link(text, link):
    return f"`{text} <{link}>`_"


def get_tool_url(toolname, version):
    if not (toolname and version):
        return ""
    return requote_uri(toolname + "_" + str(version)).lower()


def prepare_byline(
    toolid, shortdocumentation: str, contributors: List[str], versions: List[str]
):
    short_prepared = (
        shortdocumentation[:-1]
        if shortdocumentation and shortdocumentation[-1] == "."
        else shortdocumentation
    )

    contributors, versions = contributors or [], versions or []
    contributorstr = (
        f"{len(contributors)} contributor{'s' if len(contributors) != 1 else ''}"
    )
    versionstr = f"{len(versions)} version{'s' if len(versions) != 1 else ''}"
    return (
        f"``{toolid}``"
        + " · *"
        + " · ".join(t for t in [short_prepared, contributorstr, versionstr] if t)
        + "*"
    )


def nested_keys_append_with_root(d, keys: List[str], value, root_key):
    if len(keys) == 0:
        if root_key in d:
            d[root_key].append(value)
        else:
            d[root_key] = [value]
        return d
    try:
        key = keys[0]
        if key in d:
            if not isinstance(d[key], dict):
                raise NestedDictionaryTypeException(
                    key=key, error_type=type(d[key]), keys=keys
                )
            nested_keys_append_with_root(d[key], keys[1:], value, root_key=root_key)
            return d
        else:
            d[key] = nested_keys_append_with_root(
                {}, keys[1:], value, root_key=root_key
            )
    except NestedDictionaryTypeException as de:
        raise NestedDictionaryTypeException(
            key=de.key, error_type=de.error_type, keys=keys
        )

    return d


def nested_keys_add(d, keys: List[str], value) -> bool:
    if len(keys) == 0:
        raise Exception(
            f"Couldn't add {value} to nested dictionary as no keys were provided"
        )
    key = keys[0]
    if len(keys) == 1:
        if key in d:
            return False
        d[key] = value
        return True
    try:
        if key not in d:
            d[key] = {}
        elif not isinstance(d[key], dict):
            raise NestedDictionaryTypeException(
                key=key, error_type=type(d[key]), keys=keys
            )
        return nested_keys_add(d[key], keys[1:], value)
    except NestedDictionaryTypeException as de:
        raise NestedDictionaryTypeException(
            key=de.key, error_type=de.error_type, keys=keys
        )


def get_toc(title, intro_text, subpages, caption="Contents", max_depth=1):
    prepared_subpages = "\n".join(
        "   " + m.lower() for m in sorted(subpages, key=lambda l: l.lower())
    )
    return f"""
{title.replace('{title}', title)}
{"=" * len(title)}

{intro_text}

.. toctree::
   :maxdepth: {max_depth}
   :caption: {caption}:

{prepared_subpages}

*This page was auto-generated on {date.today().strftime(
        "%d/%m/%Y")}. Please do not directly alter the contents of this page.*
"""


def get_tool_toc(
    alltoolsmap: Dict[str, Dict[str, Tool]],
    title,
    intro_text,
    subpages,
    tools,
    caption="Contents",
    max_depth=1,
):

    pd = " " * 5
    mappedtools = "\n".join(
        "\n".join(pd + r for r in get_tool_row(alltoolsmap[tool]).split("\n"))
        for tool in tools
    )

    mappedmodules = "\n\n".join(
        f'     <li><a href="{m.lower()}/index.html">{m}</a></li>'
        for m in sorted(subpages)
    )

    return f"""
:orphan:

{title.replace('{title}', title)}
{"=" * len(title)}

{intro_text}

.. raw:: html

   <ul>
{mappedmodules}
   </ul>
{mappedtools}

"""


def prepare_container_warning_for_commandtool(tool: CommandTool):
    if tool.container() is not None and len(tool.container()) > 0:
        return ""

    return f"""\
.. warning::

   {tool.friendly_name()} did not include a container. You can provide one through the command line by including
   the following instruction:

   .. code-block:: bash

      janis run --container-override '{tool.id()}=<organisation/container:version>' {tool.id()}
    """


def prepare_container_warning_for_workflow(tool: Workflow):
    def recursive_find_tools_without_container(wf: Workflow):
        tools = {}
        for s in wf.step_nodes.values():
            stool = s.tool
            if isinstance(stool, Workflow):
                tools.update(recursive_find_tools_without_container(stool))
            else:
                if not stool.container():
                    tools[stool.id()] = stool

        return tools

    tools_without_containers = recursive_find_tools_without_container(tool)
    if not tools_without_containers:
        return ""

    intro = (
        "A tool in" if len(tools_without_containers) == 1 else "Some of the tools in"
    )
    merged_command = ", ".join(
        f"{t.id()}=<organisation/container:version>"
        for t in tools_without_containers.values()
    )

    return f"""\
.. warning::

   {intro} {tool.friendly_name()} did not include a container. You could provide them through
    the command line by including the following instruction:

   .. code-block:: bash

      janis run --container-override '{merged_command}' {tool.id()}
    """


def get_tool_row(tools: Dict[str, Tool]):
    distincted = {t.version(): t for t in tools.values()}
    versions = sort_tool_versions(list(distincted.keys()))
    latestversion = versions[0]

    tool = distincted[latestversion]
    meta: Metadata = tool.bind_metadata() or tool.metadata
    sd = meta.short_documentation
    sdstr = f'<p style="color: black; margin-bottom: 10px">{sd}' if sd else ""

    href = tool.id().lower() + ".html"
    return f"""\
<a href="{href}">
  <p style="margin-bottom: 5px"><b>{tool.friendly_name()}</b> <span style="margin-left: 10px; color: darkgray">{tool.id()}</span></p>
  {sdstr}
  <p><span style="margin-right: 10px; color: darkgray">({len(versions)} versions)</span>{version_html(latestversion, href=href)}</p>
</a>
<hr />
    """


def sort_tool_versions(versions: List[str]) -> List[str]:
    try:
        return sorted(versions, key=StrictVersion, reverse=True)
    except:
        return sorted(versions, reverse=True)


def version_html(version, href=None):
    fhref = f'href="{href}"' if href else ""
    return f"""\
<a class="version-button" {fhref} style="margin-bottom: 10px">
  v<b>{version[1:] if (version and version.startswith("v")) else version}</b>
</a>"""


def prepare_quickstart(tool: Tool):
    required_python_input_map = "\n".join(
        " " * 15 + i.id() + "=None,"
        for i in tool.tool_inputs()
        if not i.intype.optional
    )

    python_step_name = tool.id().lower() + "_step"
    output_python_code = "\n".join(
        " " * 7 + f'wf.output("{o.id()}", source={python_step_name}.{o.id()})'
        for o in tool.tool_outputs()
    )
    python_codeblock = f"""\
    .. code-block:: python

       from {tool.__module__} import {tool.__class__.__name__}

       wf = WorkflowBuilder("myworkflow")

       wf.step(
           "{python_step_name}",
           {tool.__class__.__name__}(
{required_python_input_map}
           )
       )
{output_python_code}
    """

    return f"""\
Quickstart
-----------

{python_codeblock}

*OR*

{prepare_run_instructions(tool)}

"""


def prepare_source(source):
    if isinstance(source, list):
        return ", ".join(prepare_source(s) for s in source)
    elif isinstance(source, str):
        return source
    elif isinstance(source, dict):
        return "\n".join(f"* {k}: {prepare_source(v)}" for k, v in source.items())
    return str(source)


def prepare_run_instructions(tool: Tool):
    metadata = tool.bind_metadata() or tool.metadata
    has_array_of_arrays_inps = True
    bla = any(
        (isinstance(i.intype, Array) and isinstance(i.intype.subtype(), Array))
        for i in tool.tool_inputs()
    )

    static_input_tuples = [
        [i.id(), i.intype.id(), prepare_source(i.doc.source)]
        for i in tool.tool_inputs()
        if i.doc.quality == InputQualityType.static and i.doc.source is not None
    ]

    reference_information = ""

    if len(static_input_tuples) > 0:
        static_input_headers = ["Name", "Type", "Source"]
        reference_information = tabulate(
            static_input_tuples, headers=static_input_headers, tablefmt="rst"
        )

    # overrides = metadata.sample_input_overrides or {}
    user_inps = {}
    other_inps = {}

    for i in tool.tool_inputs():
        if i.intype.optional or i.default is not None:
            continue

        val = i.doc.example or prepare_default_for_type(i.id(), i.intype)
        if i.doc and i.doc.quality and i.doc.quality != InputQualityType.user:
            other_inps[i.id()] = val
        else:
            # catch None and InputQualityType.user
            user_inps[i.id()] = val

    if has_array_of_arrays_inps:
        return prepare_run_instructions_input_file(
            tool, user_inps, other_inps, reference_information
        )
    else:
        return prepare_run_instructions_cli(
            tool, user_inps, other_inps, reference_information
        )


def prepare_default_for_type(identifier: str, t: DataType, idx=None):
    if isinstance(t, Array):
        st = t.subtype()
        return [
            prepare_default_for_type(identifier, st, 0),
            prepare_default_for_type(identifier, st, 1),
        ]
    elif isinstance(t, File):
        return (
            identifier
            + ("_" + str(idx) if idx is not None else "")
            + str(t.extension or "")
        )

    elif isinstance(t, String):
        return "<value>"
    elif isinstance(t, Int):
        return 0
    elif isinstance(t, Float):
        return 0.0
    elif isinstance(t, Boolean):
        return True
    return None


def prepare_run_instructions_input_file(
    tool: Tool, user_inps: dict, other_inps: dict, reference_information: str
):
    yaml_user_inps = CwlTranslator.stringify_translated_inputs(user_inps)
    yaml_other_inps = CwlTranslator.stringify_translated_inputs(other_inps)
    indented_user = "".join(" " * 7 + s for s in yaml_user_inps.splitlines(True))
    indented_other = "".join(" " * 7 + s for s in yaml_other_inps.splitlines(True))

    not_localising_secondary_warning = ""
    if isinstance(tool, WorkflowBase):
        inputs_that_arent_localising_secondary_files = [
            t.id() for t in tool.tool_inputs() if t.doc.skip_sourcing_secondary_files
        ]
        if len(inputs_that_arent_localising_secondary_files) > 0:
            not_localising_secondary_warning = f"""\
.. warning::

   The secondary files for the inputs '{"', '".join(inputs_that_arent_localising_secondary_files)}' will not automatically \
   localise using janis prepare and are built just after download. Please note this can take a few hours to build \
   before the pipeline runs. 
"""

    has_static = len(other_inps) > 0

    tb = " " * 4
    run_args = ["janis run [...run options]", tb + "--inputs inputs.yaml"]

    static_generation = (
        ""
        if not has_static
        else f"""\
   # static inputs
   janis inputs --static {tool.id()} > static.yaml"""
    )
    static_yaml = (
        ""
        if not has_static
        else f"""\
**static.yaml**

.. code-block:: yaml

{indented_other}"""
    )
    if has_static:
        run_args.append(tb + "--inputs static.yaml")

    if isinstance(tool, CommandTool) and not tool.container():
        run_args.append(
            tb + f"--container-override '{tool.id()}=<organisation/container:version>'"
        )

    run_args.append(tb + tool.id())
    run_statement = " \\\n".join(" " * 3 + el for el in run_args)

    if reference_information:
        reference_information = f"The following inputs have a suggested source. Using janis prepare with the relevant \
        ``--source-hint`` will automatically download these files. See `below <#additional-configuration-inputs>`_ for \
        more information about inputs for {tool.id()}.\n{reference_information}"

    return f"""\
1. `Install Janis </tutorials/tutorial0.html>`_

2. Ensure Janis is configured to work with Docker or Singularity.

3. Ensure all reference files are available:

4. Generate user {'and static ' if has_static else ''}input files for {tool.id()}:

.. code-block:: bash

   # user inputs
   janis inputs {"--user " if has_static else ""}{tool.id()} > inputs.yaml

{static_generation}

**inputs.yaml**

.. code-block:: yaml

{indented_user}

{static_yaml}

5. Run {tool.id()} with:

.. code-block:: bash

{run_statement}

.. note::

   You can use `janis prepare <https://janis.readthedocs.io/en/latest/references/prepare.html>`_ to improve setting up your files for this {tool.type()}. See `this guide <https://janis.readthedocs.io/en/latest/references/prepare.html>`_ for more information about Janis Prepare.

   .. code-block:: text

      OUTPUT_DIR="<output-dir>"
      janis prepare \\
          --inputs inputs.yaml \\
          --output-dir $OUTPUT_DIR \\
          {tool.id()}

      # Run script that Janis automatically generates
      sh $OUTPUT_DIR/run.sh

{indent(not_localising_secondary_warning, "   ")}


{indent(reference_information, '   ')}


"""


def prepare_run_instructions_cli(
    workflow: Tool, user_inps: dict, other_inps: dict, reference_information: str
):
    sp = 7 * " "
    cli = []
    iters = [*list(user_inps.items()), *list(other_inps.items())]
    for k, v in iters:
        vv = v
        if isinstance(v, bool):
            cli.append("--" + k)
        else:
            if isinstance(v, list):
                vv = " ".join(str(v))
            cli.append("--" + k + " " + str(vv))

    clis = (" \\\n" + sp).join(cli)

    return f"""\
1. `Install Janis </tutorials/tutorial0.html>`_

2. Ensure Janis is configured to work with Docker or Singularity.

3. Ensure all reference files are available:

.. note:: 

   More information about these inputs are available `below <#additional-configuration-inputs>`_.

{reference_information}

4. In your terminal, you can run the workflow with the following command:

.. code-block:: bash

   janis run [...workflow options] {workflow.id()} \\
       {clis}

See below for documentation for each input.
"""



import janis_core.operators.operator as gen
import janis_core.operators.standard as std
import janis_core.operators.logical as lgc

from constants import PROJECT_ROOT_DIR

sections = [
    ("Regular operators", gen),
    ("Logical operators", lgc),
    ("Standard library", std),
]


# DON'T CHANGE ANYTHING BELOW THIS LINE TO INCLUDE NEW OPERATORS

from inspect import isabstract, isclass
from janis_core import Logger
from janis_core.operators import Operator

docs_dir = PROJECT_ROOT_DIR + "/docs/"
listoperators_path = docs_dir + "references/listoperators.rst"


def prepare_operators():

    ignore = {"StringFormatter"}

    prepared_sections = []
    section_titles = []

    for title, module in sections:
        section_titles.append(title)
        prepared_sections.append(prepare_section(title, module))
        print(f"Processed operators '{title}'")

    nl = "\n"
    outp = f"""
List of operators
===================

This document is automatically generated, and contains the operators that Janis
provides. They are split up by section ({", ".join(section_titles)}) with
the relevant Python import.

For more information about expressions and operators, visit `Expressions in Janis <https://janis.readthedocs.io/en/latest/references/expressions.html>`_.

{nl.join(prepared_sections)}

*This page was autogenerated.*
"""

    with open(listoperators_path, "w+") as f:
        f.write(outp)


def prepare_section(title, module):
    prepared_operators = []

    importstatement = None

    for t, T in vars(module).items():
        if isabstract(T):
            continue
        if not isclass(T):
            continue
        if not issubclass(T, Operator):
            continue

        rettype = T.friendly_signature()

        if importstatement is None:
            importstatement = f"from {T.__module__} import *"

        prepared_operators.append(
            f"""
- **{t}**: ``{rettype}``

{T.__doc__ or ""}
"""
        )

    prepped = "\n\n".join(prepared_operators)
    return f"""
{title}
{len(title) * "+"}

Import with: ``{importstatement}``

{module.__doc__ or ""}


{prepped}
"""


if __name__ == "__main__":
    prepare_operators()



import argparse, sys, os, subprocess

from janis_core import Logger

from janisdk.container import do_container, add_container_args
from janisdk.fromcwl import do_fromcwl, add_fromcwl_args
from janisdk.fromwdl import do_fromwdl, add_fromwdl_args
from janisdk.runtest import runner as test_runner

from janis_assistant.management.configuration import JanisConfiguration


def process_args():
    cmds = {"container": do_container, "run-test": do_runtest, "fromcwl": do_fromcwl, "fromwdl": do_fromwdl}

    parser = argparse.ArgumentParser(description="Execute a workflow")
    subparsers = parser.add_subparsers(help="subcommand help", dest="command")
    parser.add_argument("-d", "--debug", action="store_true")

    subparsers.add_parser("version")
    add_container_args(subparsers.add_parser("container"))
    test_runner.add_runtest_args(subparsers.add_parser("run-test"))
    add_fromcwl_args(subparsers.add_parser("fromcwl"))
    add_fromwdl_args(subparsers.add_parser("fromwdl"))

    args = parser.parse_args()
    return cmds[args.command](args)


def do_runtest(args):
    config = None
    if args.config:
        config = JanisConfiguration.initial_configuration(path=args.config)

    runner_path = test_runner.__file__

    cli_args = sys.argv[2:]
    run_test_commands = ["python", runner_path] + cli_args

    if config:
        commands = config.template.template.prepare_run_test_command(run_test_commands)
    else:
        commands = run_test_commands

    joined_command = "' '".join(commands)
    Logger.info(f"Deploying test with command: '{joined_command}'")
    subprocess.run(commands)


if __name__ == "__main__":
    process_args()



from datetime import date
from inspect import isclass

from janis_core import DataType


def prepare_data_type(dt: DataType):
    dt_name = dt.name()
    secondary = ""

    name = dt.__name__ if isclass(dt) else dt.__class__.__name__

    if dt.secondary_files():
        mapped_secs = "\n".join(f"- ``{s}``" for s in dt.secondary_files())
        secondary = f"""\
Secondary Files
---------------

{mapped_secs}

.. note:: 

   For more information, visit `Secondary / Accessory Files <https://janis.readthedocs.io/en/latest/references/secondaryfiles.html>`__
"""

    return f"""
{dt_name}
{"=" * len(dt_name)}

{dt.doc()}

{secondary}

Quickstart
-----------

.. code-block:: python

   from {dt.__module__} import {name}

   w = WorkflowBuilder("my_workflow")

   w.input("input_{name.lower()}", {name}(optional=False))
   
   # ...other workflow steps

*This page was automatically generated on {date.today().strftime("%Y-%m-%d")}*.
"""



#!/usr/bin/env python3

"""
This class regenerates all the tool definitions

It's a bit of a random collection of things that should be refactored:

    - NestedDictionary (store values in nested structure (with root key))
    - RST Helpers
"""
from inspect import isfunction, ismodule, isabstract, isclass
from janis_assistant.templates import (
    get_all_templates,
    get_schema_for_template,
    EnvironmentTemplate,
)

from shutil import rmtree

from constants import PROJECT_ROOT_DIR
from janis_core import (
    Array,
    DataType,
    Workflow,
    CommandTool,
    Tool,
    Metadata,
    Logger,
    JanisShed,
    WorkflowMetadata,
    ToolType,
    WorkflowBase,
)

import janis_unix, janis_bioinformatics

# Import modules here so that the tool toolbox knows about them

# Output settings
from docs.generationhelpers.commandtool import prepare_commandtool_page
from docs.generationhelpers.codetool import prepare_code_tool_page
from docs.generationhelpers.datatype import prepare_data_type
from docs.generationhelpers.pipelines import (
    generate_pipeline_box,
    prepare_published_pipeline_page,
)
from docs.generationhelpers.template import prepare_template
from docs.generationhelpers.utils import (
    sort_tool_versions,
    get_tool_url,
    nested_keys_append_with_root,
    get_toc,
    get_tool_toc,
)
from docs.generationhelpers.workflow import prepare_workflow_page

docs_dir = PROJECT_ROOT_DIR + "/docs/"
tools_dir = docs_dir + "tools/"
dt_dir = docs_dir + "datatypes/"
templates_dir = docs_dir + "templates/"
pipelines_dir = docs_dir + "pipelines/"

modules = [janis_bioinformatics, janis_unix]


tool_module_information = {
    "bioinformatics": {
        "name": "Bioinformatics tools",
        "description": "A collection of tools for bioinformatics",
    },
    "unix": {
        "name": "Unix tools",
        "description": "A collection of tools available on unix systems",
    },
}

###### Shouldn't need to touch below this line #####

import os
import tabulate
from datetime import date, datetime
from typing import List, Set, Type, Tuple, Dict
import traceback


def prepare_tool(
    tool: Tool,
    toolversions: List[str],
    isorphan: bool,
    is_published_pipeline: bool = False,
):
    # Stuff to list on the documentation page:
    #   - Versions of tools
    #   - Generated command
    #   - Cool if it grouped the tools by vendor
    #   -

    if not tool:
        return None
    try:
        if is_published_pipeline:
            return ""
        if tool.type() == ToolType.CommandTool:
            return prepare_commandtool_page(tool, toolversions)
        elif tool.type() == ToolType.Workflow:
            return prepare_workflow_page(tool, toolversions)
        elif tool.type() == ToolType.CodeTool:
            return prepare_code_tool_page(tool, toolversions)
    except Exception as e:
        traceback.print_exc()
        Logger.critical(
            "Couldn't generate documentation for " + tool.id() + " " + str(e)
        )


def prepare_all_tools():
    JanisShed.hydrate(modules=[janis_unix, janis_bioinformatics])

    data_types = JanisShed.get_all_datatypes()
    tools = {
        ts[0].id(): {t.version(): t for t in ts} for ts in JanisShed.get_all_tools()
    }

    Logger.info(f"Preparing documentation for {len(tools)} tools")
    Logger.info(f"Preparing documentation for {len(data_types)} data_types")

    tool_module_index = {}
    dt_module_index = {}
    ROOT_KEY = "root"

    if os.path.exists(tools_dir):
        rmtree(tools_dir)

    for toolname, toolsbyversion in tools.items():
        # tool = tool_vs[0][0]()
        tool_versions = sort_tool_versions(list(toolsbyversion.keys()))
        default_version = tool_versions[0]
        Logger.log(
            f"Preparing {toolname}, found {len(tool_versions)} version[s] ({','.join(tool_versions)})"
        )

        defaulttool = toolsbyversion[default_version]
        if isclass(defaulttool):
            defaulttool = defaulttool()
        try:
            tool_path_components = list(
                filter(
                    lambda a: bool(a),
                    [defaulttool.tool_module(), defaulttool.tool_provider()],
                )
            )
        except Exception as e:
            Logger.critical(f"Failed to generate docs for {toolname}: {e}")
            continue

        # (toolURL, tool, isPrimary)
        toolurl_to_tool = [(toolname.lower(), defaulttool, True)] + [
            (get_tool_url(toolname, v), toolsbyversion[v], False) for v in tool_versions
        ]

        path_components = "/".join(tool_path_components)
        output_dir = f"{tools_dir}/{path_components}/".lower()
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        for (toolurl, tool, isprimary) in toolurl_to_tool:
            output_str = prepare_tool(tool, tool_versions, not isprimary)
            output_filename = output_dir + toolurl + ".rst"
            if isinstance(tool, WorkflowBase):
                tool.get_dot_plot(output_directory=output_dir, log_to_stdout=False)
            if output_str is None:
                Logger.warn(f"Skipping {tool.id()}")
                continue
            with open(output_filename, "w+") as tool_file:
                tool_file.write(output_str)

        nested_keys_append_with_root(
            tool_module_index, tool_path_components, toolname, root_key=ROOT_KEY
        )

        Logger.log("Prepared " + toolname)

    for d in data_types:
        # tool = tool_vs[0][0]()
        if issubclass(d, Array):
            Logger.info("Skipping Array DataType")
            continue
        try:
            dt = d()
        except:
            print(d.__name__ + " failed to instantiate")
            continue
        did = dt.name().lower()
        Logger.log("Preparing " + dt.name())
        output_str = prepare_data_type(dt)

        dt_path_components = []
        # dt_path_components = list(filter(
        #     lambda a: bool(a),
        #     [, tool.tool_provider()]
        # ))

        path_components = "/".join(dt_path_components)
        output_dir = f"{dt_dir}{path_components}/"
        output_filename = output_dir + did + ".rst"

        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        nested_keys_append_with_root(
            dt_module_index, dt_path_components, did, root_key=ROOT_KEY
        )

        with open(output_filename, "w+") as dt_file:
            dt_file.write(output_str)

        Logger.log("Prepared " + did)

    def prepare_modules_in_index(contents, title, dir, max_depth=1):
        module_filename = dir + "/index.rst"
        module_tools = sorted(set(contents[ROOT_KEY] if ROOT_KEY in contents else []))
        submodule_keys = sorted(m for m in contents.keys() if m != ROOT_KEY)
        indexed_submodules_tools = [m.lower() for m in submodule_keys]

        with open(module_filename, "w+") as module_file:
            module_file.write(
                get_tool_toc(
                    alltoolsmap=tools,
                    title=title,
                    intro_text=f"Automatically generated index page for {title}:",
                    subpages=indexed_submodules_tools,
                    tools=module_tools,
                    max_depth=max_depth,
                )
            )

        for submodule in submodule_keys:
            prepare_modules_in_index(
                contents=contents[submodule], title=submodule, dir=f"{dir}/{submodule}/"
            )

    def prepare_dtmodules_in_index(contents, title, dir, max_depth=1):
        module_filename = dir + "/index.rst"
        module_tools = sorted(set(contents[ROOT_KEY] if ROOT_KEY in contents else []))
        submodule_keys = sorted(m for m in contents.keys() if m != ROOT_KEY)
        indexed_submodules_tools = [m.lower() + "/index" for m in submodule_keys]

        with open(module_filename, "w+") as module_file:
            module_file.write(
                get_toc(
                    title=title,
                    intro_text=f"Automatically generated index page for {title}:",
                    subpages=indexed_submodules_tools + module_tools,
                    max_depth=max_depth,
                )
            )

        for submodule in submodule_keys:
            prepare_modules_in_index(
                contents=contents[submodule], title=submodule, dir=f"{dir}/{submodule}/"
            )

    prepare_modules_in_index(tool_module_index, title="Tools", dir=tools_dir)
    prepare_dtmodules_in_index(
        dt_module_index, title="Data Types", dir=dt_dir, max_depth=1
    )


def prepare_templates():
    """
    :return:
    """

    os.makedirs(templates_dir, exist_ok=True)

    templates = get_all_templates()

    for tkey, template in templates.items():
        with open(os.path.join(templates_dir, tkey + ".rst"), "w+") as f:
            f.write(prepare_template(tkey, template))

    introtext = """\

Templates
###########
    
These templates are used to configure Cromwell / CWLTool broadly. For more information, visit `Configuring Janis <https://janis.readthedocs.io/en/latest/references/configuration.html#cromwell>`__.
"""

    with open(os.path.join(templates_dir, "index.rst"), "w+") as f:
        f.write(
            introtext
            + get_toc(
                title="",
                intro_text="List of templates for ``janis-assistant``:",
                subpages=list(templates.keys()),
            )
        )


def generate_pipelines_page():
    import janis_pipelines

    print("Generating pipelines page...")

    modules = [janis_pipelines]
    workflows: List[Workflow] = []

    for module in modules:

        workflows.extend(
            cls()
            for n, cls in list(module.__dict__.items())
            if not n.startswith("__")
            and type(cls) != type
            and isclass(cls)
            and hasattr(cls, "type")
            and cls.type() == ToolType.Workflow
        )

    wf_strings = "\n".join(
        generate_pipeline_box(w, leading_space=" " * 8) for w in workflows
    )

    page = f"""\
Pipelines
=========

.. raw:: html

    <div id="box" style="display: flex; flex-flow: wrap;">
{wf_strings}
    </div>
"""

    os.makedirs(pipelines_dir, exist_ok=True)

    with open(os.path.join(pipelines_dir, "index.rst"), "w+") as f:
        f.write(page)

    # Write all the pages
    for w in workflows:
        toolstr = prepare_published_pipeline_page(w, [w.version()])
        w.get_dot_plot(output_directory=pipelines_dir, log_to_stdout=False)

        with open(os.path.join(pipelines_dir, w.id().lower() + ".rst"), "w+") as f:
            f.write(toolstr)


if __name__ == "__main__":
    prepare_all_tools()
    prepare_templates()
    generate_pipelines_page()



from typing import List

from janis_core import (
    ToolOutput,
    ToolInput,
    WorkflowBuilder,
    File,
    Array,
    ScatterDescription,
    ScatterMethods,
)
from janis_core.types import InputSelector
from janis_core.utils.scatter import ScatterMethod
from janis_unix.tools.unixtool import UnixTool
from janis_bioinformatics.data_types import FastaBwa, BamBai


class DataTypeWithSecondary(File):
    def __init__(self, optional=False):
        super().__init__(optional, extension=".ext")

    @staticmethod
    def name():
        return "DataTypeWithSecondary"

    @staticmethod
    def secondary_files():
        return ["^.txt"]


class ToolThatAcceptsAndReturnsSecondary(UnixTool):
    @staticmethod
    def tool():
        return "ToolThatAcceptsAndReturnsSecondary"

    def friendly_name(self) -> str:
        return "Example tool only"

    @staticmethod
    def base_command():
        return "echo"  # non functional tool

    def inputs(self) -> List[ToolInput]:
        return [ToolInput("inp", DataTypeWithSecondary())]

    def outputs(self) -> List[ToolOutput]:
        return [ToolOutput("out", DataTypeWithSecondary(), glob=InputSelector("inp"))]


class ToolTypeThatAcceptsMultipleBioinfTypes(UnixTool):
    def tool(self):
        return "TESTTOOL_BIOINF"

    def inputs(self):
        return [ToolInput("bam", BamBai), ToolInput("reference", FastaBwa)]

    def outputs(self):
        return [
            ToolOutput("out_bam", BamBai, glob=InputSelector("bam")),
            ToolOutput("out_reference", FastaBwa, glob=InputSelector("reference")),
        ]

    def base_command(self):
        return "echo"


if __name__ == "__main__":
    w = WorkflowBuilder("test_workflow")

    # EXAMPLE 1

    w.input("inp", DataTypeWithSecondary)
    w.step("stp", ToolThatAcceptsAndReturnsSecondary(inp=w.inp))
    w.output("out", source=w.stp)
    w.translate("wdl")

    # EXAMPLE 2

    w2 = WorkflowBuilder("scattered_test_workflow")
    w2.input("inp", Array(DataTypeWithSecondary), value=["path/to/file.ext"])
    w2.step("stp", ToolThatAcceptsAndReturnsSecondary(inp=w2.inp), scatter="inp")
    w2.output("out", source=w2.stp)
    w2.translate("wdl")

    # EXAMPLE 3

    w3 = WorkflowBuilder("scattered_bioinf_complex")
    w3.input("my_bams", Array(BamBai))
    w3.input("my_references", Array(FastaBwa))

    w3.step(
        "my_step",
        ToolTypeThatAcceptsMultipleBioinfTypes(
            bam=w3.my_bams, reference=w3.my_references
        ),
        scatter=ScatterDescription(["bam", "reference"], ScatterMethods.cross),
    )

    w3.output("out_bam", source=w3.my_step.out_bam)
    w3.output("out_reference", source=w3.my_step.out_reference)

    w3.translate("wdl")



"""
simple.py

    Date: 2019-05-03
    Author: Michael Franklin


    This class represents a simple unix workflow, and was prepared as part of the
    following tutorial: https://janis.readthedocs.io/en/latest/tutorials/simple.html

    This version of the simple workflow IS wrapped in a subclass of Workflow to
    demonstrate how a subworkflow can be subclassed and easily be used again.

    Notes:
        - The toolshed only accepts workflows that are a subclass of `Workflow`,
        - It's easier to associate metadata with a subclass,
        - ** The runner (currently titled Shepherd) has a much easier time finding
                a subclass over a variable of type.


    Workflow description:
        1. A file is untarred
        2. The result of the untar is compiled
        3. The (untarred + compiled) result is tarred into a new archive

        The following ASCII diagram is a basic visual representation of the DAG:

        file.tar → untar → compile → tar -> out.tar
                          ↘_______↗
"""

# The classes we require to build a basic workflow
from janis_core import Workflow

# Data types - These help us logically connect workflows
from janis.data_types import TarFile

# Tools - The command line tools we're going to call
from janis.tools import Compile, Untar


class SimpleWorkflow(Workflow):
    def id(self) -> str:
        return "simple"

    def friendly_name(self):
        return "Simple workflow: Untar, Compile, Tar"

    def constructor(self):
        self.input("tarfile", TarFile)
        self.step("untar", Untar(tarfile=self.tarfile))
        self.step("compile", Compile(file=self.untar.out), scatter="file")
        self.output("out", source=self.compile)


if __name__ == "__main__":
    SimpleWorkflow().translate("wdl", to_disk=True)



from typing import Optional, Union, List, Tuple

from janis_core.utils.validators import Validators

from .cltconvert import convert_command_tool_fragments
from janis_core import (
    ToolMetadata,
    String,
    Logger,
    JanisShed,
    DataType,
    Boolean,
    Filename,
    CommandToolBuilder,
)
from janis_core.tool.commandtool import ToolInput

from .templates import ToolTemplateType

container_exec = {"docker": ["docker", "run"], "singularity": ["singularity", "exec"]}


common_replacements = {"input": "inp", "output": "outputFilename"}

option_markers = {
    "options:",
    "arguments:",
    "required arguments:",
    "optional arguments:",
}


def get_help_from_container(
    container: str,
    basecommand: Union[str, List[str]],
    help_param: Optional[str] = "--help",
    containersoftware="docker",
):
    import subprocess, os

    bc = basecommand if isinstance(basecommand, list) else [basecommand]

    if help_param:
        bc = [*bc, help_param]

    cmd = [*container_exec[containersoftware]]

    if containersoftware == "docker":
        cmd.extend([container, *bc])
    elif containersoftware == "singularity":
        # this doesn't work: https://github.com/hpcng/singularity/issues/5316
        cmd.extend(["docker://" + container, *bc])

    print("Running command: " + " ".join(f"'{x}'" for x in cmd))
    try:
        help = (
            subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            .decode("utf-8")
            .rstrip()
        )
    except subprocess.CalledProcessError as e:
        help = str(e.output)
    return help


def get_version_from_container(
    container: str,
    basecommand: Union[str, List[str]],
    versionparam: Optional[str] = "--version",
    containersoftware="docker",
):
    import subprocess

    bc = basecommand if isinstance(basecommand, list) else [basecommand]
    cmd = [*container_exec[containersoftware], container, *bc, versionparam]

    print("Running command: " + " ".join(f"'{x}'" for x in cmd))
    try:
        help = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        return None

    return help.decode("utf-8").rstrip()


def first_or_default(iterable, default=None):
    filtered = [f for f in iterable if f is not None]
    if len(filtered) > 0:
        return filtered[0]
    return default


def parse_str(
    helpstr, option_marker: str = None, requires_prev_line_blank_or_param=False
):
    doc = ""
    args = []
    lines = helpstr.replace("\\n", "\n").split("\n")
    options_idx = None
    markers = option_markers
    if option_marker:
        markers = markers.union({option_marker.lower()})
    for il in range(len(lines)):
        line = lines[il]
        if not line.lstrip():
            continue

        ll = line.strip().lower()

        if any(ll.startswith(m) for m in markers):
            options_idx = il
            break

        doc += line + "\n"

    if options_idx is None:
        raise Exception("Couldn't find the start of the inputs")

    prev_arg = None
    last_line_was_blank_or_param = True

    while options_idx < len(lines) - 1:
        options_idx += 1

        line = lines[options_idx]

        if not line.lstrip():
            # line is empty
            prev_arg = None
            last_line_was_blank_or_param = True
            continue

        line_args = [l.strip() for l in line.lstrip().split("  ") if l]
        largs = len(line_args)

        if largs == 0:
            raise Exception("No args when should have been filtered by previous step")

        tool_doc = ""

        if (
            not requires_prev_line_blank_or_param or last_line_was_blank_or_param
        ) and line_args[0].startswith("-"):
            # sometimes this section has two items
            processed_tags = [
                get_tag_and_cleanup_prefix(p) for p in line_args[0].split(",")
            ]
            processed_tags = [t for t in processed_tags if t is not None]
            if len(processed_tags) < 1:
                continue
            tags = sorted(processed_tags, key=lambda l: len(l[1]), reverse=True)
            potential_type = first_or_default([p[3] for p in processed_tags])

            if len(tags) > 1:
                tool_doc += "(" + ", ".join(t[0] for t in tags[1:]) + ") "

            if largs > 1:
                tool_doc += " ".join(line_args[1:])

            prefix, tag, has_equal, guessed_type = tags[0]
            eqifrequired = "=" if has_equal else ""

            if not potential_type:
                potential_type = Boolean

            if len(tag) == 1:
                while not Validators.validate_identifier(tag):
                    print(
                        f"The tag for '{prefix}' was invalid, we need you to come up with a new identifier for:"
                    )
                    print("\t" + tool_doc if tool_doc else line)
                    tag = str(input("New identifier: "))
            try:
                prev_arg = ToolInput(
                    tag,
                    potential_type(optional=True),
                    prefix=prefix + eqifrequired,
                    separate_value_from_prefix=not has_equal,
                    doc=tool_doc.replace('"', "'"),
                )
            except:
                print(f"Skipping '{tag}' as it wasn't validated correctly")
            args.append(prev_arg)
            # we'll get the longer one for the tag

        elif prev_arg:
            prev_arg.doc.doc += " " + line.lstrip()
        else:
            last_line_was_blank_or_param = False

    return doc, args


def guess_type(potential_type: str):
    if not potential_type:
        return None
    l = potential_type.lower()
    hopeful_type = JanisShed.get_datatype(l)

    if not hopeful_type:
        if "st" in potential_type:
            hopeful_type = String

    if hopeful_type:
        Logger.info(f"Found type {hopeful_type.__name__} from tag: {potential_type}")

    return hopeful_type


def get_tag_and_cleanup_prefix(
    prefix,
) -> Optional[Tuple[str, str, bool, Optional[DataType]]]:
    """
    :param prefix:
    :return: (raw_element, potentialID, hasSeparator, potentialType)
    """
    # cases:
    # -a ADAPTER
    # --adapter=ADAPTER
    # --quality-cutoff=[5'CUTOFF,]3'CUTOFF
    el = prefix.lstrip()
    has_equals = False
    pretag = None
    potential_type = None

    # if prefix is split by ':' or split by
    if ":" in el or "=" in el:
        parts = None
        if ":" in el:
            parts = el.split(":")
        elif "=" in el:
            parts = el.split("=")
            has_equals = True

        if len(parts) > 2:
            Logger.warn(
                f"Unexpected number of components in the tag '{el}' to guess the type, using '{parts[0]}' and skipping type inference"
            )
        else:
            el, pt = parts[0], guess_type(parts[1])

            if not potential_type and pt:
                potential_type = pt

    if " " in el:
        el = el.split(" ")[0]

    titleComponents = [l.strip().lower() for l in el.split("-") if l]
    if len(titleComponents) == 0:
        Logger.critical(
            f"Title components for tag '{prefix}' does not have a component"
        )
        return None
    tag = "_".join(titleComponents)

    if tag.lower() in common_replacements:
        tag = common_replacements[tag.lower()]

    if tag.lower() == "outputfilename":
        potential_type = Filename

    return el, tag, has_equals, potential_type


def from_container(
    container: str,
    basecommand: List[str],
    helpcommand="-h",
    containersoftware="docker",
    optionsmarker: Optional[str] = None,
    name: Optional[str] = None,
    version: Optional[str] = None,
    type: ToolTemplateType = ToolTemplateType.base,
):
    helpstr = get_help_from_container(
        container=container,
        basecommand=basecommand,
        help_param=helpcommand,
        containersoftware=containersoftware,
    )
    tooldoc, args = parse_str(helpstr, option_marker=optionsmarker)

    if not version:
        comps = container.split(":")
        version = comps[-1] if len(comps) > 1 else "Latest"

    if not version:
        version = get_version_from_container(
            container,
            basecommand,
            versionparam="-v",
            containersoftware=containersoftware,
        )

    tool_id = name or basecommand
    if isinstance(tool_id, list):
        tool_id = "".join(s.title() for s in tool_id)
    else:
        tool_id = tool_id[0].upper() + tool_id[1:]

    t = CommandToolBuilder(
        tool=tool_id,
        base_command=basecommand,
        inputs=args,
        outputs=[],
        metadata=ToolMetadata(documentation=tooldoc) if tooldoc else None,
        version=version,
        container=container,
    )

    return t.translate("janis", to_console=False), helpstr

    # return (
    #     convert_command_tool_fragments(
    #         type=type,
    #         toolid=name or basecommand,
    #         basecommand=basecommand,
    #         friendly_name="".join(basecommand),
    #         toolprov="TOOLPROVIDER",
    #         ins=args,
    #         outs=[],
    #         metadata=ToolMetadata(documentation=tooldoc),
    #         container=container,
    #         version=version,
    #     ),
    #     helpstr,
    # )



from textwrap import indent
from typing import List

from janis_core import WorkflowMetadata, Workflow
from requests.utils import requote_uri
from tabulate import tabulate

from docs.generationhelpers.utils import prepare_run_instructions
from .utils import prepare_byline, format_rst_link, get_tool_url, version_html


def generate_pipeline_box(workflow: Workflow, leading_space=""):

    meta: WorkflowMetadata = workflow.bind_metadata() or workflow.metadata

    tag_component = lambda tag: f'<span class="no-select tagelement">{tag}</span>'
    href = f"{workflow.id().lower()}.html"

    tags = "".join(tag_component(t) for t in (meta.keywords or []))
    # date: datetime = meta.dateUpdated or meta.dateCreated

    contributors = meta.contributors or []
    max_display_conts = 5
    if len(contributors) < max_display_conts:
        contributorstr = ", ".join(meta.contributors or ["None"])
    else:
        nothers = len(contributors) - max_display_conts + 2
        contributorstr = (
            ", ".join(meta.contributors[: max_display_conts - 2])
            + f" and {nothers} others"
        )

    return "\n".join(
        leading_space + l
        for l in f"""
<div class="col-6" style="margin: 10px; padding: 20px; border: 1px solid #e3e3e3; border-radius: 5px;">
    <h4 style="margin-bottom: 10px"><a href="{href}">{workflow.friendly_name()}</a></h4>
    {f'<p style="overflow-x: scroll;">{tags}</p>' if tags else ""}
    <p>{meta.short_documentation or "<em>Short documentation required</em>"}</p>
    <p style="margin-bottom: 10px; font-size: 12px; margin-top: 0px;">Contributors: {contributorstr}</p>
    {version_html(workflow.version() or meta.version)}
</div>
""".split(
            "\n"
        )
    )


def prepare_published_pipeline_page(workflow: Workflow, versions: List[str]):
    if not workflow:
        return None

    # tool_modules = tool.__module__.split(".") # janis._bioinformatics_.tools.$toolproducer.$toolname.$version

    metadata: WorkflowMetadata = workflow.bind_metadata() or workflow.metadata

    if not workflow.friendly_name():
        raise Exception(
            f"Tool '{type(workflow).__name__}' ({workflow.id()}) did not provide the required 'friendly_name' for the docs"
        )

    fn = workflow.friendly_name() if workflow.friendly_name() else workflow.id()
    en = f" ({workflow.id()})" if fn != workflow.id() else ""
    tn = fn + en

    onelinedescription = prepare_byline(
        workflow.id(), metadata.short_documentation, metadata.contributors, versions
    )

    citation = "\n\n".join([el for el in [metadata.citation, metadata.doi] if el])

    toolmetadata = [
        ("ID", f"``{workflow.id()}``"),
        ("Versions", ", ".join(str(s) for s in versions[::-1]) if versions else ""),
        ("Authors", ", ".join(metadata.contributors)),
        ("Citations", citation),
        ("Created", str(metadata.dateCreated)),
        ("Updated", str(metadata.dateUpdated)),
    ]

    cwl = workflow.translate("cwl", to_console=False, allow_empty_container=True)[0]
    wdl = workflow.translate("wdl", to_console=False, allow_empty_container=True)[0]

    formatted_url = (
        format_rst_link(metadata.documentationUrl, metadata.documentationUrl)
        if metadata.documentationUrl
        else "*No URL to the documentation was provided*"
    )

    embeddedtoolsraw = {
        f"{s.tool.id()}/{s.tool.version()}": s.tool
        for s in workflow.step_nodes.values()
    }
    embeddedtools = tabulate(
        [
            [tool.friendly_name(), f"``{key}``"]
            for key, tool in embeddedtoolsraw.items()
        ],
        tablefmt="rst",
    )

    input_headers = ["name", "type", "documentation"]

    required_input_tuples = [
        [i.id(), i.intype.id(), i.doc.doc]
        for i in workflow.tool_inputs()
        if not i.intype.optional
    ]
    optional_input_tuples = [
        [i.id(), i.intype.id(), i.doc.doc]
        for i in workflow.tool_inputs()
        if i.intype.optional
    ]

    formatted_inputs = tabulate(
        required_input_tuples + optional_input_tuples, input_headers, tablefmt="rst"
    )

    formatted_toolversions_array = []
    formatted_toolincludes_array = []
    for v in versions:
        link = get_tool_url(workflow.id(), v)
        formatted_toolincludes_array.append(".. include:: " + link)
        if v == workflow.version():
            formatted_toolversions_array.append(
                f"- {v} (current)"
            )  # + format_rst_link(v + " (current)", link))
        else:
            formatted_toolversions_array.append(
                "- " + format_rst_link(v, link + ".html")
            )

    output_headers = ["name", "type", "documentation"]
    output_tuples = [
        [o.id(), o.outtype.id(), o.doc.doc] for o in workflow.tool_outputs()
    ]
    formatted_outputs = tabulate(output_tuples, output_headers, tablefmt="rst")

    tool_prov = ""
    if workflow.tool_provider() is None:
        print("Tool :" + workflow.id() + " has no company")
    else:
        tool_prov = "." + workflow.tool_provider().lower()

    workflow_image = requote_uri(workflow.versioned_id()) + ".dot.png"

    nl = "\n"

    return f"""\
:orphan:

{fn}
{"=" * len(tn)}

{onelinedescription}

{metadata.documentation if metadata.documentation else "No documentation was provided: " + format_rst_link(
    "contribute one", f"https://github.com/PMCC-BioinformaticsCore/janis-{workflow.tool_module()}")}

Quickstart
-----------

{prepare_run_instructions(workflow)}

Outputs
-----------

{formatted_outputs}

Workflow
--------

.. image:: {workflow_image}


Information
------------


{nl.join(f":{key}: {value}" for key, value in toolmetadata)}

Embedded Tools
~~~~~~~~~~~~~~~~~

{embeddedtools}


Additional configuration (inputs)
---------------------------------

{formatted_inputs}

Workflow Description Language
------------------------------

.. code-block:: text

{indent(wdl, "   ")}

Common Workflow Language
-------------------------

.. code-block:: text

{indent(cwl, "   ")}
"""



"""
simple.py

    Date: 2019-10-03
    Author: Michael Franklin


    This file contains a simple unix workflow, and was prepared as part of the
    following tutorial: https://janis.readthedocs.io/en/latest/tutorials/simple.html

    This version of the simple workflow is NOT wrapped in a subclass of Workflow.

    Workflow description:
        1. A file is untarred
        2. The result of the untar is compiled
        3. The (untarred + compiled) result is tarred into a new archive

        The following ASCII diagram is a basic visual representation of the DAG:

        file.tar → untar → compile → tar -> out.tar
                          ↘_______↗
"""

# The classes we require to build a basic workflow
from janis import WorkflowBuilder
from janis.data_types import TarFile
from janis.tools import Compile, Untar

w = WorkflowBuilder("user_guide")

w.input("tarfile", TarFile, value="/path/to/file.tar")
w.step("untar", Untar(tarfile=w.tarfile))
w.step("compile", Compile(file=w.untar), scatter="file")
w.output("compiled", source=w.compile.out)

if __name__ == "__main__":
    w.translate("wdl", to_disk=True, validate=True)



from textwrap import indent
from typing import List
from tabulate import tabulate


from janis_core import CommandTool, ToolMetadata

from docs.generationhelpers.utils import (
    prepare_quickstart,
    prepare_container_warning_for_commandtool,
)
from .utils import prepare_byline, format_rst_link, get_tool_url


def prepare_commandtool_page(tool: CommandTool, versions: List[str]):
    if not tool:
        return None

    # tool_modules = tool.__module__.split(".") # janis._bioinformatics_.tools.$toolproducer.$toolname.$version

    metadata: ToolMetadata = tool.bind_metadata() or tool.metadata

    if not tool.friendly_name():
        raise Exception(
            f"Tool '{type(tool).__name__}' ({tool.id()}) did not provide the required 'friendly_name' for the docs"
        )

    fn = tool.friendly_name() if tool.friendly_name() else tool.id()
    en = f" ({tool.id()})" if fn != tool.id() else ""
    tn = fn + en

    has_container = tool.container() is not None

    onelinedescription = prepare_byline(
        tool.id(), metadata.short_documentation, metadata.contributors, versions
    )

    formatted_url = (
        format_rst_link(metadata.documentationUrl, metadata.documentationUrl)
        if metadata.documentationUrl
        else "*No URL to the documentation was provided*"
    )

    toolmetadata = [
        t
        for t in [
            ("ID", f"``{tool.id()}``"),
            ("URL", formatted_url),
            ("Versions", ", ".join(versions)),
            ("Container", tool.container()),
            ("Authors", ", ".join(metadata.contributors)),
            ("Citations", metadata.citation),
            ("DOI", metadata.doi) if metadata.doi else None,
            ("Created", str(metadata.dateCreated)),
            ("Updated", str(metadata.dateUpdated)),
        ]
        if t and len(t) == 2
    ]

    input_headers = ["name", "type", "prefix", "position", "documentation"]
    argument_headers = ["value", "prefix", "position", "documentation"]

    required_input_tuples = [
        [i.id(), i.input_type.id(), i.prefix, i.position, i.doc.doc if i.doc else ""]
        for i in tool.inputs()
        if not i.input_type.optional
    ]
    optional_input_tuples = [
        [i.id(), i.input_type.id(), i.prefix, i.position, i.doc.doc if i.doc else ""]
        for i in tool.inputs()
        if i.input_type.optional
    ]

    formatted_args = None
    args = tool.arguments()
    if args:
        argument_tuples = [
            [str(a.value), a.prefix, a.position, a.doc.doc if a.doc else ""]
            for a in tool.arguments()
        ]
        fargs = tabulate(argument_tuples, headers=argument_headers, tablefmt="rst")
        formatted_args = "Arguments\n----------\n\n" + fargs

    formatted_inputs = tabulate(
        required_input_tuples + optional_input_tuples, input_headers, tablefmt="rst"
    )

    formatted_toolversions_array = []
    formatted_toolincludes_array = []
    for v in versions:
        link = get_tool_url(tool.id(), v)
        formatted_toolincludes_array.append(".. include:: " + link)
        if v == tool.version():
            formatted_toolversions_array.append(
                f"- {v} (current)"
            )  # + format_rst_link(v + " (current)", link))
        else:
            formatted_toolversions_array.append(
                "- " + format_rst_link(v, link + ".html")
            )

    output_headers = ["name", "type", "documentation"]
    output_tuples = [[o.id(), o.outtype.id(), o.doc] for o in tool.tool_outputs()]
    formatted_outputs = tabulate(output_tuples, output_headers, tablefmt="rst")

    cwl = tool.translate("cwl", to_console=False, allow_empty_container=True)
    wdl = tool.translate("wdl", to_console=False, allow_empty_container=True)

    tool_prov = ""
    if tool.tool_provider() is None:
        print("Tool :" + tool.id() + " has no company")
    else:
        tool_prov = "." + tool.tool_provider().lower()

    nl = "\n"

    return f"""\
:orphan:

{fn}
{"=" * len(tn)}

{onelinedescription}

{metadata.documentation if metadata.documentation else "No documentation was provided: " + format_rst_link(
    "contribute one", f"https://github.com/PMCC-BioinformaticsCore/janis-{tool.tool_module()}")}

{prepare_container_warning_for_commandtool(tool)}
{prepare_quickstart(tool)}

Information
------------

{nl.join(f":{key}: {value}" for key, value in toolmetadata)}


Outputs
-----------

{formatted_outputs}


Additional configuration (inputs)
---------------------------------

{formatted_inputs}

Workflow Description Language
------------------------------

.. code-block:: text

{indent(wdl, "   ")}

Common Workflow Language
-------------------------

.. code-block:: text

{indent(cwl, "   ")}

"""



from typing import Dict

from janis_core import DynamicWorkflow, Array, String
from janis_unix.tools import Echo


class MyFirstDynamicWorkflow(DynamicWorkflow):
    def friendly_name(self):
        return "My first dynamic workflow"

    def id(self) -> str:
        return "MyFirstDynamicWorkflow"

    def constructor(self, inputs: Dict[str, any], hints: Dict[str, str]):

        my_value_to_print = inputs.get("inp")

        if isinstance(my_value_to_print, list):
            # print each element if 'inp' is a list
            self.input("inp_list", Array(String))
            self.step("print", Echo(inp=self.inp_list), scatter="inp")
        else:
            # print only once
            self.input("inp", String)
            self.step("print", Echo(inp=self.inp))

        # Janis will derive the output from the source (Array(String) | String)
        self.output("out", source=self.print)

    def modify_inputs(self, inputs, hints: Dict[str, str]):
        my_value_to_print = inputs.get("inp")
        if isinstance(my_value_to_print, list):
            return {**inputs, "inp_list": my_value_to_print}
        else:
            return inputs



from textwrap import dedent

import ruamel.yaml
from tabulate import tabulate

from janis_assistant.templates import EnvironmentTemplate, get_schema_for_template


def prepare_template(name: str, template: EnvironmentTemplate):
    schema = get_schema_for_template(template)

    tname = name.replace("_", " ").title()

    required_params = [s for s in schema if not s.optional]

    required_cli_params = "".join(
        " \\\n       --" + s.id() + ("" if s.type == bool else " <value>")
        for s in required_params
    )

    required_yaml = {}
    for s in required_params:
        if s.type == bool:
            required_yaml[s.id()] = True
        else:
            required_yaml[s.id()] = "<value>"
    required_yaml_str = (
        ""
        if not required_yaml
        else "".join(
            " " * 5 + s
            for s in ruamel.yaml.dump(
                required_yaml, default_flow_style=False
            ).splitlines(True)
        )
    )

    optional = [(s.id(), s.type, s.default, s.doc or "") for s in schema if s.optional]

    # mapped

    required_section = ""

    if required_params:
        required = [(s.id(), s.type, s.doc or "") for s in required_params]
        requiredf = tabulate(required, ["ID", "Type", "Documentation"], tablefmt="rst")
        required_section = f"""\
**Required**

{requiredf}"""

    optionalf = tabulate(
        optional, ["ID", "Type", "Default", "Documentation"], tablefmt="rst"
    )

    template_doc = dedent(template.__doc__ or "")

    return f"""\
{tname}
{"=" * len(tname)}

Template ID: ``{name}``

{template_doc}

Quickstart
-----------

Take note below how to configure the template. This quickstart only includes the fields you absolutely require. \
This will write a new configuration to ``~/.janis.conf``. See `configuring janis <https://janis.readthedocs.io/en/latest/references/configuration.html>`__ for more information.

.. code-block:: bash

   janis init {name}{required_cli_params}
   
   # or to find out more information
   janis init {name} --help

OR you can insert the following lines into your template:

.. code-block:: yaml

   template:
     id: {name}
{required_yaml_str}


Fields
-------

{required_section}

**Optional**

{optionalf}

"""
